/*
 *
 *  Copyright (c) 2004-2005 Warren Jasper <wjasper@tx.ncsu.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 ************************
 * This has been substantially modified by Nick Glazzard, January 2013.
 * Primarily to consistently return status and remove unused code.
 ************************
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <syslog.h>
#include <asm/types.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <linux/hiddev.h>
#include "pmd.h"

// Driver Functions

int PMD_Find(int vendor, int product, int fd[], int maxifs)
//---------------------------------------------------------
// Locate (device,product) on /dev/hiddev* devices.
// For whatever reason, this finds the same device on all the hiddev* devices it tries.
// (Ubuntu 12.04 LTS).
{
  int i;
  int fd_temp;
  int num_interfaces = 0;    // number of interfaces for this HID device
  char devname[80];
  char devname_old[80];
  struct hiddev_devinfo device_info;

  for(i = 0; i < MAX_HIDDEV; i++){
    sprintf(devname, "/dev/hiddev%d", i);         // new locatin for hiddev
    sprintf(devname_old, "/dev/usb/hiddev%d", i); // old location for hiddev
    if ((fd_temp = open(devname, O_RDONLY)) < 0) {
      if ((fd_temp = open(devname_old, O_RDONLY)) < 0) {
        continue;
      }
    }
    /* get device information */
    if ( ioctl(fd_temp, HIDIOCGDEVINFO, &device_info) >= 0) {
      /* check vendor and product */
      printf( "i=%d, vendor=0x%x, product=0x%x\n", i, device_info.vendor, device_info.product );
      if (vendor == device_info.vendor && product == device_info.product) {
	/* this is us */
        fd[num_interfaces++] = fd_temp;
	if( num_interfaces == maxifs )
	  break;
	continue;
      }
    }
    close(fd_temp);
  }
  return num_interfaces;
}

int PMD_SendOutputReport(int fd, __u8 reportID, __u8* vals, int num_vals, int delay)
//----------------------------------------------------------------------------------
// Write data to device. Return -1 on error.
{
  int ret = 0, i;
  struct hiddev_report_info out_report;
  struct hiddev_usage_ref uref;

  uref.report_type = HID_REPORT_TYPE_OUTPUT;
  uref.report_id = reportID;
  uref.field_index = 0;
  uref.usage_index = 0;

  /* usage code for this for this usage */
  ret = ioctl(fd, HIDIOCGUCODE, &uref);
  if (ret < 0) {
    perror("SendCommandFS: ioctl to get usage code for out report");
    return ret;
  }

  /* fill in usage values. */
  for( i = 0; (i < num_vals) && vals; i++){
    uref.usage_index = i;
    uref.value = vals[i];
    ret = ioctl(fd, HIDIOCSUSAGE, &uref);
    if (ret < 0) {
      perror("SendCommandFS:ioctl to set usage value for out report");
      return ret;
    }
  }

  /* tell the driver about the usage values */
  out_report.report_type = HID_REPORT_TYPE_OUTPUT;
  out_report.report_id = reportID;
  ioctl(fd, HIDIOCGREPORTINFO, &out_report);
  if (0 > ret) {
    perror("SendCommandFS: ioctl to get out report info");
    return ret;
  }

  /* this ioctl puts the report on the wire */
  ret = ioctl(fd, HIDIOCSREPORT, &out_report);
  if (0 > ret) {
    perror("SendCommandFS: ioctl to send output report");
    return ret;
  }

  usleep(delay);
  return ret;
}


int PMD_GetInputReport(int fd, __u8 reportID, __u8 *vals, int num_vals, int delay)
//--------------------------------------------------------------------------------
// Read data from device. Return -1 on error
{
  int ret = 0, i;
  struct hiddev_usage_ref uref;

  uref.report_type = HID_REPORT_TYPE_INPUT;
  uref.report_id = reportID;
  uref.field_index = 0;
  uref.usage_index = 0;
  uref.usage_code = 0xff000001;

  for( i = 0; i < num_vals; i++){
    uref.usage_index = i;
    ret = ioctl(fd, HIDIOCGUSAGE, &uref);
    usleep(delay);
    if(0 > ret){
      perror("ioctl to set usage value for in report");
      return ret;
    }
    vals[i] = (__u8)(uref.value & 0x000000FF) ;
  }
  return ret;
}


int PMD_GetFeatureReport(int fd, __u8 reportID, __u8 *vals, int num_vals, int delay)
//----------------------------------------------------------------------------------
// Query device (?) Return -1 on error.
{
  int ret = 0, i;
  struct hiddev_usage_ref uref;

  uref.report_type = HID_REPORT_TYPE_FEATURE;
  uref.report_id = reportID;
  uref.field_index = 0;
  uref.usage_index = 0;

  ret = ioctl(fd, HIDIOCGUCODE, &uref);
  if( 0 > ret) {
    perror("ioctl to get feature report usage code");
    return ret;
  }

  for( i = 0; i < num_vals; i++){
    uref.usage_index = i;
    ret = ioctl(fd, HIDIOCGUSAGE, &uref);
    usleep(delay);
    if(0 > ret){
      perror("ioctl to set usage value for feature report");
      return ret;
    }
    vals[i] = (__u8)(uref.value & 0x000000FF) ;
  }
  return ret;
}

