/*
 *
 *  Copyright (c) 2004-2005  Warren Jasper <wjasper@tx.ncsu.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 ************************
 * This has been substantially modified by Nick Glazzard, January 2013.
 * Several things have just been deleted. In particular, the original
 * analog input calibration scheme (which just would not work with my
 * PMD-1608FS, for unknown reasons -- EEPROM values seemed always wrong),
 * multi-channel input functions, firmware update functions.
 * A completely different calibration scheme has been implemented.
 ************************
 */

/*********************************/
/* PMD-1608FS SPECIFIC FUNCTIONS */
/*********************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <asm/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <linux/hiddev.h>

#include "pmd.h"
#include "usb-1608FS.h"

#define FS_DELAY 1000


/****************************/
/* ANALOG INPUT CALIBRATION */
/****************************/

int usbCalibrationIO( Calibration_AIN table_AIN[NGAINS_1608FS][NCHAN_1608FS], char* fname, int save )
//---------------------------------------------------------------------------------------------------
// Save or get a calibration table to/from file fname.
// The table must be generated at some point by: usbCalibrate_USB1608FS() so it can be saved.
{
  FILE* f;
  if( save ){
    f = fopen( fname, "wb+" );
    if( f != NULL ){
      time_t curtime;
      time( &curtime );
      if( fwrite( &curtime, sizeof(curtime), 1, f ) != 1 ){
	fprintf( stderr, "e1\n" );
	fclose( f );
	return( 0 );
      }
      if( fwrite( table_AIN, (size_t)(NGAINS_1608FS*NCHAN_1608FS*sizeof(Calibration_AIN)), 1, f ) != 1 ){
	fprintf( stderr, "e2\n" );
	fclose( f );
	return( 0 );
      }
      fclose( f );
      return( 1 );
    }
    else{
      fprintf( stderr, "e3\n" );
      return( 0 );
    }
  }
  else{
    f = fopen( fname, "rb" );
    if( f != NULL ){
      time_t savedtime;
      if( fread( &savedtime, sizeof(savedtime), 1, f ) != 1 ){
	fprintf( stderr, "e4\n" );
	fclose( f );
	return( 0 );
      }
      printf( "Calibration data saved on: %s", ctime( &savedtime ) );
      if( fread( table_AIN, (size_t)(NGAINS_1608FS*NCHAN_1608FS*sizeof(Calibration_AIN)), 1, f ) != 1 ){
	fprintf( stderr, "e5\n" );
	fclose( f );
	return( 0 );
      }
      fclose( f );
      return( 1 );
    }
    else{
      fprintf( stderr, "e6\n" );
      return( 0 );
    }
  }
}

int usbCalibrate_USB1608FS(int fd, Calibration_AIN table_AIN[NGAINS_1608FS][NCHAN_1608FS],
			   int slow, float* known_calvolts)
//---------------------------------------------------------------------------------------------------
// Calibrate the analog inputs. To use this, connect together pins: 1,3,5,7,9,11,13,15,17.
// 17 is the cal voltage output. The others are the 8 analog inputs. You may want to measure
// the actual cal voltages periodically to verify they have not changed. Set slow to 1 for that.
// If known_calvolts is not NULL it should be a table of 5 voltages measured for each cal voltage
// at pin 17.
//
// InstaCAL (Windows utility from MCC) should do this and store values in EEPROM in the 1608FS,
// but that does not work for me. Hence the original BuildCalTable does not work. (Nan's everywhere).
//
// Return 0 on error, else 1.
{
  int i, j, k;
  float* calvolts;
#define NUM_VALS 5
  // calvolts is the accurately measured actual output voltage from
  // the cal terminal for each possible cal output setting.
  // Nominally:        0.0     0.625   1.25    2.5     5.0
  float nominal_calvolts[] = {-0.002f, 0.621f, 1.244f, 2.491f, 4.981f };
  float calvalues[NUM_VALS];
  int ncalused[] = {5,4,3,2};
  int gainstab[] = {BP_10_00V, BP_5_00V, BP_2_00V, BP_1_00V};

  if( known_calvolts != NULL )
    calvolts = known_calvolts;
  else
    calvolts = nominal_calvolts;
  useconds_t wait = 50000;
  if( slow )wait *= 200;

  // Step over known gain settings.
  for( k=0; k<NGAINS_1608FS; k++ ){
    int cur_gain = gainstab[k];
    int ncals = ncalused[k];

    // Step over channels
    for( i=0; i<8; i++ ){
      printf( "CHAN %d\n", i );
      // Measure the raw A/D values for each useful cal voltage output.
      for( j=0; j<ncals; j++ ){
	if( ! usbSetCal_USB1608FS(fd, j) )
	  return 0;
	usleep(50000);
	calvalues[j] = (float)usbRawAIn_USB1608FS(fd, i, cur_gain) - 32768.0f;
      }
      // Linear regression: Fit straight line through raw value versus known cal voltage.
      float xbar = 0.0f;
      float ybar = 0.0f;
      for( j=0; j<ncals; j++ ){
	xbar += calvolts[j];
	ybar += calvalues[j];
      }
      xbar /= ncals;
      ybar /= ncals;
      float sumtop = 0.0f;
      float sumbot = 0.0f;
      for( j=0; j<ncals; j++ ){
	sumtop += (calvolts[j] - xbar) * (calvalues[j] - ybar);
	sumbot += (calvolts[j] - xbar) * (calvolts[j] - xbar);
      }
      float beta = sumtop / sumbot;
      float alpha = ybar - beta * xbar;
      printf( "Slope = %f, Offset = %f\n", beta, alpha );
      // Verify that gives sensible results converting cal voltages.
      for( j=0; j<ncals; j++ ){
	if( ! usbSetCal_USB1608FS(fd, j) )
	  return 0;
	usleep(50000);
	float invalue = (float)usbRawAIn_USB1608FS(fd, i, cur_gain) - 32768.0f;
	float volts = ( invalue - alpha ) / beta;
	printf( "Invalue = %.1f, Cal = %f, CalVal = %.1f, Measured = %f\n", 
		invalue, calvolts[j], calvalues[j], volts );
      }
      if( ! usbSetCal_USB1608FS(fd, 0) )
	return 0;
      table_AIN[k][i].slope = 1.0f / beta;     // slope
      table_AIN[k][i].offset = alpha;   // intercept
    }
    printf( "========\n" );
  }
  return( 1 );
}

int usbSetCal_USB1608FS(int fd, __u8 value )
//-------------------------------------------
// Set the calibration output voltage to [value].
// Nominal outputs for value: 0:0.0, 1:0.625, 2:1.25, 3:2.5, 4:5.0
{
  __u8 cmd[1];

  if( value >= 0 && value < 5 ){
    cmd[0] = value; 
    if( PMD_SendOutputReport(fd, SET_CAL, cmd, sizeof(cmd), FS_DELAY) < 0 )
      return 0;
    return 1;
  }
  else
    return 0;
}


/*********************/
/* DIGITAL I/O LINES */
/*********************/

int usbDConfigPort_USB1608FS(int fd, __u8 direction)
//---------------------------------------------------
// AFAICT, sets ALL 8 digital I/O lines to be input or output.
// [direction] = DIO_DIR_IN or DIO_DIR_OUT.
// Then usbDIn_USB1608FS() & usbDOut_USB1608FS() read/write
// an entire 8 bit word from/to the DIO lines.
{
  __u8 cmd[1];
  cmd[0] = direction;
  if( PMD_SendOutputReport(fd, DCONFIG, cmd, sizeof(cmd), FS_DELAY) < 0 )
    return 0;
  return 1;
}

int usbDConfigBit_USB1608FS(int fd, __u8 bit_num, __u8 direction)
//----------------------------------------------------------------
// AFAICT, set a single DIO line (0:7?) to be DIO_DIR_IN or DIO_DIR_OUT.
{
  __u8 cmd[2];
  cmd[0] = bit_num;
  cmd[1] = direction;
  if( PMD_SendOutputReport(fd, DCONFIG_BIT, cmd, sizeof(cmd), FS_DELAY) < 0 )
    return 0;
  return 1;
}

int usbDIn_USB1608FS(int fd, __u8* value)
//----------------------------------------
// Read 8 bits to value from the 8 DIO lines (set as input).
{
  if( PMD_SendOutputReport(fd, DIN, 0, 0, FS_DELAY) < 0 )
    return 0;
  if( PMD_GetInputReport(fd, DIN, value, sizeof(*value), FS_DELAY) < 0 )
    return 0;
  return 1;
}

int usbDInBit_USB1608FS(int fd, __u8 bit_num, __u8* value)
//---------------------------------------------------------
// Read a single DIO line (0:7) to value (0 or 1, I think).
{
  __u8 cmd[1];
  cmd[0] = bit_num;
  if( PMD_SendOutputReport(fd, DBIT_IN, cmd, sizeof(cmd), FS_DELAY) < 0 )
    return 0;
  if( PMD_GetInputReport(fd, DBIT_IN, value, sizeof(*value), FS_DELAY) < 0 )
    return 0;
  return 1;
}

int usbDOut_USB1608FS(int fd, __u8 value)
//----------------------------------------
// Write 8 bits from value to the DIO lines (set as output).
{
  __u8 cmd[1];
  cmd[0] = value;
  if( PMD_SendOutputReport(fd, DOUT, cmd, sizeof(cmd), FS_DELAY) < 0 )
    return 0;
  return 1;
}

int usbDOutBit_USB1608FS(int fd, __u8 bit_num, __u8 value)
//---------------------------------------------------------
// Set or clear DIO line bit_num (0:7) according to value (0:1).
{
  __u8 cmd[2];

  cmd[0] = bit_num;
  cmd[1] = value;

  if( PMD_SendOutputReport(fd, DBIT_OUT, cmd, sizeof(cmd), FS_DELAY) < 0 )
    return 0;
  return 1;
}


/****************/
/* ANALOG INPUT */
/****************/

unsigned short usbRawAIn_USB1608FS(int fd, __u8 channel, __u8 range)
//------------------------------------------------------------------
// Read a raw 16 bit value from input [channel] using gain [range].
// This is used to generate the calibration table.
{
  __u8 cmd[2];
  __u16 data;

  cmd[0] = channel;
  cmd[1] = range;
  
  if (channel > NCHAN_1608FS - 1) {
    printf("usbRawAIn_USB1608FS: channel out of range.\n");
    return -1;
  }

  if (range > 7) {
    printf("usbRawAIn_USB1608FS: range setting too large.\n");
    return -1;
  }

  PMD_SendOutputReport(fd, AIN, cmd, sizeof(cmd), FS_DELAY);
  data = 0x0;
  usleep(1000);
  PMD_GetInputReport(fd, AIN, (__u8 *) &data, sizeof(data), FS_DELAY);
  
  //printf("Value (raw) = 0x%x\n", data ); fflush(stdout);
  return( data );
}

float usbAIn_USB1608FS(int fd, __u8 channel, __u8 range, Calibration_AIN table_AIN[NGAINS_1608FS][NCHAN_1608FS])
//--------------------------------------------------------------------------------------------------------------
// Read the voltage on [channel] with gain setting [range], using calibration data in table_AIN.
// The result is in volts.
{
  __u8 cmd[2];
  __u16 data;
  float value = 0.0f;
  float escale = 1.0f;

  cmd[0] = channel;
  cmd[1] = range;
  
  if (channel > NCHAN_1608FS - 1) {
    printf("usbAIn_USB1608FS: channel out of range.\n");
    return -9999999.9f;
  }

  if (range > 7) {
    printf("usbAIN: range setting too large.\n");
    return -9999999.9f;
  }

  // Get raw input value.
  if( PMD_SendOutputReport(fd, AIN, cmd, sizeof(cmd), FS_DELAY) < 0 )
    return -9999999.9f;
    
  data = 0x0;
  usleep(1000);
  if( PMD_GetInputReport(fd, AIN, (__u8 *) &data, sizeof(data), FS_DELAY) )
    return -9999999.9f;

  // Get an extra scale or ranges that don't have direct calibration data.
  switch( range ){
  case BP_2_50V: escale = 2.5f / 2.0f; break;
  case BP_1_25V: escale = 1.25f; break;
  case BP_0_625V: escale = 0.625f; break;
  case BP_0_3125V: escale = 0.3125f; break;
  }

  // Convert the raw data value to a calibrated voltage reading.
  switch (range) {
    case  BP_10_00V:
      //printf( "slope=%f, offset=%f\n", table_AIN[0][channel].slope, table_AIN[0][channel].offset);
      value = table_AIN[0][channel].slope * (((float) data - 32768.0f) - table_AIN[0][channel].offset);
      break;
    case BP_5_00V:
      //printf( "slope=%f, offset=%f\n", table_AIN[0][channel].slope, table_AIN[0][channel].offset);
      value = table_AIN[1][channel].slope * (((float) data - 32768.0f) - table_AIN[1][channel].offset);
      break;
    case BP_2_50V:  // use data from 2 V
    case BP_2_00V:
      //printf( "slope=%f, offset=%f\n", table_AIN[0][channel].slope, table_AIN[0][channel].offset);
      value = escale * table_AIN[2][channel].slope * (((float) data - 32768.0f) + table_AIN[2][channel].offset);
      break;
    case BP_1_25V:  // use data from 1 V
    case BP_1_00V:
    case BP_0_625V:
    case BP_0_3125V:
      //printf( "slope=%f, offset=%f\n", table_AIN[0][channel].slope, table_AIN[0][channel].offset);
      value = escale * table_AIN[3][channel].slope * (((float) data - 32768.0f) + table_AIN[3][channel].offset);
      break;
    default:
      break;
  }

  return value;
}


/***********/
/* COUNTER */
/***********/

int usbInitCounter_USB1608FS(int fd)
//-----------------------------------
// Reset the counter to 0.
{
  if( PMD_SendOutputReport(fd, CINIT, 0, 0, FS_DELAY) < 0 )
    return 0;
  return 1;
}

__u32 usbReadCounter_USB1608FS(int fd)
//------------------------------------
// Read the current counter count.
{
  __u32 value;

  PMD_SendOutputReport(fd, CIN, 0, 0, FS_DELAY);
  PMD_GetInputReport(fd, CIN, (__u8  *) &value, sizeof(value), FS_DELAY);

  return value;
}


/************/
/* SUNDRIES */
/************/

int usbBlink_USB1608FS(int fd)
//----------------------------
// Make the LED on the device blink.
{
  if( PMD_SendOutputReport(fd, BLINK_LED, 0, 0, FS_DELAY) < 0 )
    return 0;
  return 1;
}

int usbReset_USB1608FS(int fd)
//----------------------------
// Reset the USB connection to the device.
{
  if( PMD_SendOutputReport(fd, RESET, 0, 0, FS_DELAY) < 0 )
    return 0;
  return 1;
}

__u16 usbGetStatus_USB1608FS(int fd)
//----------------------------------
// Return the status of the 1608FS. Unfortunately, I don't know what these status
// bits mean ...
{
  __u16 status;

  PMD_SendOutputReport(fd, GET_STATUS, 0, 0, FS_DELAY);
  PMD_GetInputReport(fd, GET_STATUS, (__u8 *) &status, sizeof(status), FS_DELAY);
  return status;
}

int usbReadCode_USB1608FS(int fd, __u32 address, __u8 count, __u8 data[])
//-----------------------------------------------------------------------
// Read from the device's firmware Flash memory.
{
  struct t_arg {
    __u8 address[3];
    __u8 count;     // 62 max.
  } arg;

  if ( count > 62 ) count = 62;  

  memcpy(&arg.address[0], &address, 3);   // 24 bit address
  arg.count = count;
  if( PMD_SendOutputReport(fd, READ_CODE, (__u8 *) &arg, sizeof(arg), FS_DELAY) < 0 )
    return 0;
  if( PMD_GetInputReport(fd, READ_CODE,  data, count, FS_DELAY) < 0 )
    return 0;
  return 1;
}

int usbReadSerial_USB1608FS( int fd, char serial[9] )
//----------------------------------------------------
// Read the serial number of the device.
{
  int i;
  int bRead;

  for( i=0; i<8; i++ )serial[i] = ' ';
  bRead = usbReadCode_USB1608FS(fd, 0x200000, 8, (__u8*) serial);
  serial[8] = '\0';
  printf("Serial Number = \"%s\"\n", serial);
  return bRead;
}
